#ifndef MAPPER_H
#define MAPPER_H

#define ROM_MAXPAGES	0x200 /* max 8KB pages */
#define ROM_MAXSIZE	(ROM_MAXPAGES*0x2000)
#define ROMPAGE_DUMMY	(ROM_MAXPAGES+2)

enum {
	CARTTYPE_ASCII8=0,
	CARTTYPE_ASCII8S,
	CARTTYPE_ASCII16,
	CARTTYPE_ASCII16S,
	CARTTYPE_BTL80,
	CARTTYPE_BTL90,
	CARTTYPE_BTL126,
	CARTTYPE_CROSSBLAIM,
	CARTTYPE_RTYPE,
	CARTTYPE_KOEI,
	CARTTYPE_KONAMI,
	CARTTYPE_GAMEMASTER2,
	CARTTYPE_KONAMISCC,
	CARTTYPE_KONAMISCCI,
	CARTTYPE_KONAMISYN,
	CARTTYPE_MFLASHSCC,
	CARTTYPE_MATRAINK,
	CARTTYPE_HARRYFOX,
	CARTTYPE_PAC,
	CARTTYPE_MIRROR,
	CARTTYPE_START0000,
	CARTTYPE_START4000,
	CARTTYPE_START8000,
	CARTTYPE_STARTC000,
	CARTTYPE_PLAYBALL,
	CARTTYPE_DSK2ROM,
	CARTTYPE_ZEMINAZMB,
	
	CARTTYPE_MAX
};

#define MCF_SRAM	1
#define MCF_EMPTY	2

typedef __fastcall u8(*fp_mapread)(u16);
fp_mapread mapread[8]; /* 64KB (8*8KB) Z80 memmap */
typedef __fastcall void(*fp_mapwrite)(u16,u8);
fp_mapwrite mapwrite[8];

u8 __fastcall mapreadnull(u16);		void __fastcall mapwritenull(u16,u8);
u8 __fastcall mapreadbios(u16);		u8 __fastcall mapreadbioshack(u16);
u8 __fastcall mapreadram(u16);		void __fastcall mapwriteram(u16,u8);

u8 __fastcall mapreadc10(u16);		u8 __fastcall mapreadc20(u16);
u8 __fastcall mapreadc12(u16);		u8 __fastcall mapreadc22(u16);
u8 __fastcall mapreadc14(u16);		u8 __fastcall mapreadc24(u16);
u8 __fastcall mapreadc16(u16);		u8 __fastcall mapreadc26(u16);
u8 __fastcall mapreadc18(u16);		u8 __fastcall mapreadc28(u16);
u8 __fastcall mapreadc1a(u16);		u8 __fastcall mapreadc2a(u16);
u8 __fastcall mapreadc1c(u16);		u8 __fastcall mapreadc2c(u16);
u8 __fastcall mapreadc1e(u16);		u8 __fastcall mapreadc2e(u16);


void mapper_init(void);
void mapper_clean(void);
void mapper_flush_ram(void);
void mapper_set_buffer_counter(int);
void mapper_sound_stream(signed short*,int);
void mapper_new_frame(void);
void mapper_end_frame(void);

void __fastcall mapper_write_pslot(u8);
u8 __fastcall mapper_read_pslot(void);

void __fastcall mapper_refresh_cb(int);
void __fastcall mapper_refresh_pslot_read(void);
void __fastcall mapper_refresh_pslot_write(void);

void mapper_open_bios(const char*);
void mapper_update_bios_hack(void);
int mapper_open_cart(int,const char*,int);
void mapper_close_cart(int);
void mapper_init_cart(int);
void mapper_set_carttype(int,u32);
int mapper_get_carttype(int);
u32 mapper_get_cartcrc(int);
u32 mapper_get_bioscrc(void);
int mapper_get_autodetect_crc(void);
int mapper_get_cartsize(int);
u32 mapper_autodetect_cart(const char*);

int mapper_get_type_uid(u32);
int mapper_get_uid_type(u32);
const char* mapper_get_type_longname(u32);
const char* mapper_get_type_shortname(u32);
u32 mapper_get_type_flags(u32);
char* mapper_get_current_name(char*);
const char* mapper_get_file(int);
const char* mapper_get_defaultbios_file(void);
const char* mapper_get_bios_file(void);
const char* mapper_get_bios_name(void);

int __fastcall mapper_state_get_version(void);
int __fastcall mapper_state_get_size(void);
void __fastcall mapper_state_save(u8**);
void __fastcall mapper_state_load_cur(u8**);
int __fastcall mapper_state_load(int,u8**);

#endif /* MAPPER_H */
