/* ASCII mappers (and similar) */
/* depending on the game, ASCII mapper size is usually larger than the ROM size,
 unpredictable without a complete database or header */


/* --- ASCII 8
	used by many non-Konami megaROMs, eg.
	1942, Batman, Daiva Story 4, Dragon Quest 1,2, Dragon Slayer 4, Fairy Land Story,
	Fantasm Soldier Valis, Fantasy Zone, Mitsume ga Tooru, Super Laydock
*/
void __fastcall mapwrite_ascii8(u16 a,u8 v)
{
	/* all pages switchable at $6000-$7fff */
	int slot=~pslot>>2&1;
	
	cartbank[slot][(a>>11&3)+2]=v&cartmask[slot];
	refresh_cb(slot);
}

static void mapinit_ascii8(int slot)
{
	int i;
	
	/* 4 pages of 8KB at $4000-$bfff, init to 0 */
	for (i=2;i<6;i++) cartbank[slot][i]=0;
	
	mapwrite_std[slot+1][3]=mapwrite_ascii8;
}



/* --- ASCII 8 + SRAM (8KB)
	Deep Dungeon 1,2, Psychic War, Syougun, Xanadu
*/
static int a8s_is_sram[2][4];

u8 __fastcall mapread_ascii8s_sram_s1(u16 a) { return cartsram[0][a&0x1fff]; }
u8 __fastcall mapread_ascii8s_sram_s2(u16 a) { return cartsram[1][a&0x1fff]; }
const fp_mapread mapread_a8s[2]={mapread_ascii8s_sram_s1,mapread_ascii8s_sram_s2};

void __fastcall mapwrite_ascii8s_sram_s1(u16 a,u8 v) { cartsram[0][a&0x1fff]=v; }
void __fastcall mapwrite_ascii8s_sram_s2(u16 a,u8 v) { cartsram[1][a&0x1fff]=v; }
const fp_mapwrite mapwrite_a8s[2]={mapwrite_ascii8s_sram_s1,mapwrite_ascii8s_sram_s2};

static __inline__ void a8s_refresh(int slot)
{
	/* map sram */
	
	/* $4000-$7fff: read-only */
	if (a8s_is_sram[slot][0])	mapread_std[slot+1][2]=mapread_a8s[slot];
	else				mapread_std[slot+1][2]=lut_mapread[slot][2];
	if (a8s_is_sram[slot][1])	mapread_std[slot+1][3]=mapread_a8s[slot];
	else				mapread_std[slot+1][3]=lut_mapread[slot][3];
	
	/* $8000-$bfff: read/write */
	if (a8s_is_sram[slot][2]) {	mapread_std[slot+1][4]=mapread_a8s[slot];	mapwrite_std[slot+1][4]=mapwrite_a8s[slot];	}
	else {				mapread_std[slot+1][4]=lut_mapread[slot][4];	mapwrite_std[slot+1][4]=mapwritenull;		}
	if (a8s_is_sram[slot][3]) {	mapread_std[slot+1][5]=mapread_a8s[slot];	mapwrite_std[slot+1][5]=mapwrite_a8s[slot];	}
	else {				mapread_std[slot+1][5]=lut_mapread[slot][5];	mapwrite_std[slot+1][5]=mapwritenull;		}
	
	mapper_refresh_pslot_read();
	mapper_refresh_pslot_write();
}

void __fastcall mapwrite_ascii8s(u16 a,u8 v)
{
	/* sram */
	int slot=~pslot>>2&1;
	a8s_is_sram[slot][a>>11&3]=v&(cartmask[slot]+1);
	a8s_refresh(slot);
	
	/* standard a8 */
	mapwrite_ascii8(a,v);
}

static void mapclean_ascii8s(int slot)
{
	sram_save(slot,cartsram[slot],0x2000,NULL);
}

/* state			size
banks sram enabled		4*1

==				4
*/
#define STATE_VERSION_A8S	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_A8S		4

static int __fastcall mapgetstatesize_ascii8s(int slot)
{
	return STATE_SIZE_A8S;
}

static void __fastcall mapsavestate_ascii8s(int slot,u8** s)
{
	STATE_SAVE_1(a8s_is_sram[slot][0]);
	STATE_SAVE_1(a8s_is_sram[slot][1]);
	STATE_SAVE_1(a8s_is_sram[slot][2]);
	STATE_SAVE_1(a8s_is_sram[slot][3]);
}

static void __fastcall maploadstatecur_ascii8s(int slot,u8** s)
{
	STATE_LOAD_1(a8s_is_sram[slot][0]);
	STATE_LOAD_1(a8s_is_sram[slot][1]);
	STATE_LOAD_1(a8s_is_sram[slot][2]);
	STATE_LOAD_1(a8s_is_sram[slot][3]);
	
	a8s_refresh(slot);
}

static int __fastcall maploadstate_ascii8s(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_A8S:
			maploadstatecur_ascii8s(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_ascii8s(int slot)
{
	int i;
	
	if (mapclean[slot]==mapclean_nothing) {
		sram_load(slot,&cartsram[slot],0x2000,NULL,0xff);
		mapclean[slot]=mapclean_ascii8s;
	}
	
	mapgetstatesize[slot]=mapgetstatesize_ascii8s;
	mapsavestate[slot]=mapsavestate_ascii8s;
	maploadstatecur[slot]=maploadstatecur_ascii8s;
	maploadstate[slot]=maploadstate_ascii8s;
	
	for (i=0;i<4;i++) a8s_is_sram[slot][i]=FALSE;
	a8s_refresh(slot);
	
	/* rest of init similar to ASCII 8 */
	mapinit_ascii8(slot);
	mapwrite_std[slot+1][3]=mapwrite_ascii8s;
}



/* --- Koei (very similar to ASCII 8 + SRAM, but 32KB)
	Genghis Khan, Nobunaga no Yabou - Zenkokuhan, Sangokusi 1
	
	(though it is known that some cartridges have less than 32KB, it works fine)
*/
static int koei_is_sram[2][4];
static int koei_sram_bank[2][4];

static const int lut_koei_sb[4]={0,0x2000,0x4000,0x6000};

u8 __fastcall mapread_koei_sram_s1(u16 a) { return cartsram[0][(a&0x1fff)|koei_sram_bank[0][(a>>13)-2]]; }
u8 __fastcall mapread_koei_sram_s2(u16 a) { return cartsram[1][(a&0x1fff)|koei_sram_bank[1][(a>>13)-2]]; }
const fp_mapread mapread_koei[2]={mapread_koei_sram_s1,mapread_koei_sram_s2};

void __fastcall mapwrite_koei_sram_s1(u16 a,u8 v) { cartsram[0][(a&0x1fff)|koei_sram_bank[0][(a>>13)-2]]=v; }
void __fastcall mapwrite_koei_sram_s2(u16 a,u8 v) { cartsram[1][(a&0x1fff)|koei_sram_bank[1][(a>>13)-2]]=v; }
const fp_mapwrite mapwrite_koei[2]={mapwrite_koei_sram_s1,mapwrite_koei_sram_s2};

static __inline__ void koei_refresh(int slot)
{
	/* map sram */
	
	/* $4000-$7fff: read-only */
	if (koei_is_sram[slot][0])	mapread_std[slot+1][2]=mapread_koei[slot];
	else				mapread_std[slot+1][2]=lut_mapread[slot][2];
	if (koei_is_sram[slot][1])	mapread_std[slot+1][3]=mapread_koei[slot];
	else				mapread_std[slot+1][3]=lut_mapread[slot][3];
	
	/* $8000-$bfff: read/write */
	if (koei_is_sram[slot][2]) {	mapread_std[slot+1][4]=mapread_koei[slot];	mapwrite_std[slot+1][4]=mapwrite_koei[slot];	}
	else {				mapread_std[slot+1][4]=lut_mapread[slot][4];	mapwrite_std[slot+1][4]=mapwritenull;		}
	if (koei_is_sram[slot][3]) {	mapread_std[slot+1][5]=mapread_koei[slot];	mapwrite_std[slot+1][5]=mapwrite_koei[slot];	}
	else {				mapread_std[slot+1][5]=lut_mapread[slot][5];	mapwrite_std[slot+1][5]=mapwritenull;		}
	
	mapper_refresh_pslot_read();
	mapper_refresh_pslot_write();
}

void __fastcall mapwrite_koei_std(u16 a,u8 v)
{
	/* sram */
	int am=a>>11&3;
	int slot=~pslot>>2&1;
	
	koei_is_sram[slot][am]=v&(cartmask[slot]+1);
	koei_sram_bank[slot][am]=lut_koei_sb[v&3];
	koei_refresh(slot);
	
	/* standard a8 */
	mapwrite_ascii8(a,v);
}

static void mapclean_koei(int slot)
{
	sram_save(slot,cartsram[slot],0x8000,NULL);
}

/* state			size
banks sram enabled		4*1
sram bank numbers		4*2

==				12
*/
#define STATE_VERSION_KOEI	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_KOEI		12

static int __fastcall mapgetstatesize_koei(int slot)
{
	return STATE_SIZE_KOEI;
}

static void __fastcall mapsavestate_koei(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_SAVE_1(koei_is_sram[slot][i]);
		STATE_SAVE_2(koei_sram_bank[slot][i]);
	}
}

static void __fastcall maploadstatecur_koei(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_LOAD_1(koei_is_sram[slot][i]);
		STATE_LOAD_2(koei_sram_bank[slot][i]);
	}
	
	koei_refresh(slot);
}

static int __fastcall maploadstate_koei(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_KOEI:
			maploadstatecur_koei(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_koei(int slot)
{
	int i;
	
	if (mapclean[slot]==mapclean_nothing) {
		sram_load(slot,&cartsram[slot],0x8000,NULL,0xff);
		mapclean[slot]=mapclean_koei;
	}
	
	mapgetstatesize[slot]=mapgetstatesize_koei;
	mapsavestate[slot]=mapsavestate_koei;
	maploadstatecur[slot]=maploadstatecur_koei;
	maploadstate[slot]=maploadstate_koei;
	
	for (i=0;i<4;i++) koei_is_sram[slot][i]=FALSE;
	koei_refresh(slot);
	
	/* rest of init similar to ASCII 8 */
	mapinit_ascii8(slot);
	mapwrite_std[slot+1][3]=mapwrite_koei_std;
}



/* --- ASCII 16
	(mostly) the rest of the non-Konami megaROMs, eg.
	Gall Force, Golvellius, Jagur, Meikyuu Shinwa (Eggerland 2), Romancia, Super Pierrot (Do! Run Run)
*/
void __fastcall mapwrite_ascii16(u16 a,u8 v)
{
	if (!(a&0x800)) {
		/* pages switchable at $6000-$67ff and $7000-$77ff */
		int slot=~pslot>>2&1;
		int page=(a>>11&2)+2;
		
		cartbank[slot][page++]=v<<1&cartmask[slot];
		cartbank[slot][page]=(v<<1&cartmask[slot])|1;
		refresh_cb(slot);
	}
}

static void mapinit_ascii16(int slot)
{
	/* 2 pages of 16KB at $4000-$bfff, init to 0 */
	cartbank[slot][2]=cartbank[slot][4]=0;
	cartbank[slot][3]=cartbank[slot][5]=1;
	
	mapwrite_std[slot+1][3]=mapwrite_ascii16;
}



/* --- ASCII 16 + SRAM (2KB)
	Harryfox MSX Special, Hydlide 2
*/
static int a16s_is_sram[2][2];

u8 __fastcall mapread_ascii16s_sram_s1(u16 a) { return cartsram[0][a&0x7ff]; }
u8 __fastcall mapread_ascii16s_sram_s2(u16 a) { return cartsram[1][a&0x7ff]; }
const fp_mapread mapread_a16s[2]={mapread_ascii16s_sram_s1,mapread_ascii16s_sram_s2};

void __fastcall mapwrite_ascii16s_sram_s1(u16 a,u8 v) { cartsram[0][a&0x7ff]=v; }
void __fastcall mapwrite_ascii16s_sram_s2(u16 a,u8 v) { cartsram[1][a&0x7ff]=v; }
const fp_mapwrite mapwrite_a16s[2]={mapwrite_ascii16s_sram_s1,mapwrite_ascii16s_sram_s2};

static __inline__ void a16s_refresh(int slot)
{
	/* map sram */
	
	/* $4000-$7fff: read-only */
	if (a16s_is_sram[slot][0]) {
		mapread_std[slot+1][2]=mapread_std[slot+1][3]=mapread_a16s[slot];
	}
	else {
		mapread_std[slot+1][2]=lut_mapread[slot][2];
		mapread_std[slot+1][3]=lut_mapread[slot][3];
	}
	
	/* $8000-$bfff: read/write */
	if (a16s_is_sram[slot][1]) {
		mapread_std[slot+1][4]=mapread_std[slot+1][5]=mapread_a16s[slot];
		mapwrite_std[slot+1][4]=mapwrite_std[slot+1][5]=mapwrite_a16s[slot];
	}
	else {
		mapread_std[slot+1][4]=lut_mapread[slot][4];
		mapread_std[slot+1][5]=lut_mapread[slot][5];
		mapwrite_std[slot+1][4]=mapwrite_std[slot+1][5]=mapwritenull;
	}
	
	mapper_refresh_pslot_read();
	mapper_refresh_pslot_write();
}

void __fastcall mapwrite_ascii16s(u16 a,u8 v)
{
	if (!(a&0x800)) {
		int slot=~pslot>>2&1;
		
		/* sram */
		a16s_is_sram[slot][a>>12&1]=v&(cartmask[slot]+1);
		a16s_refresh(slot);
		
		/* standard a16 */
		mapwrite_ascii16(a,v);
	}
}

static void mapclean_ascii16s(int slot)
{
	sram_save(slot,cartsram[slot],0x800,NULL);
}

/* state			size
banks sram enabled		2*1

==				2
*/
#define STATE_VERSION_A16S	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_A16S		2

static int __fastcall mapgetstatesize_ascii16s(int slot)
{
	return STATE_SIZE_A16S;
}

static void __fastcall mapsavestate_ascii16s(int slot,u8** s)
{
	STATE_SAVE_1(a16s_is_sram[slot][0]);
	STATE_SAVE_1(a16s_is_sram[slot][1]);
}

static void __fastcall maploadstatecur_ascii16s(int slot,u8** s)
{
	STATE_LOAD_1(a16s_is_sram[slot][0]);
	STATE_LOAD_1(a16s_is_sram[slot][1]);
	
	a16s_refresh(slot);
}

static int __fastcall maploadstate_ascii16s(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_A16S:
			maploadstatecur_ascii16s(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_ascii16s(int slot)
{
	if (mapclean[slot]==mapclean_nothing) {
		sram_load(slot,&cartsram[slot],0x800,NULL,0xff);
		mapclean[slot]=mapclean_ascii16s;
	}
	
	mapgetstatesize[slot]=mapgetstatesize_ascii16s;
	mapsavestate[slot]=mapsavestate_ascii16s;
	maploadstatecur[slot]=maploadstatecur_ascii16s;
	maploadstate[slot]=maploadstate_ascii16s;
	
	a16s_is_sram[slot][0]=a16s_is_sram[slot][1]=FALSE;
	a16s_refresh(slot);
	
	/* rest of init similar to ASCII 16 */
	mapinit_ascii16(slot);
	mapwrite_std[slot+1][3]=mapwrite_ascii16s;
}
