/* miscellaneous mappers */

/* --- dB-Soft Cross Blaim (single game) */
void __fastcall mapwrite_crossblaim(u16 a,u8 v)
{
	/* 2nd page switchable at $0000-$ffff */
	int slot=~pslot>>2&1;
	
	if (v&2) {
		/* 32KB ROM 1 */
		cartbank[slot][0]=cartbank[slot][1]=cartbank[slot][6]=cartbank[slot][7]=ROMPAGE_DUMMY;
		cartbank[slot][4]=v<<1&7&cartmask[slot];
		cartbank[slot][5]=(v<<1&7&cartmask[slot])|1;
	}
	else {
		/* 32KB ROM 0, fixed, and mirrored to page 0 and page 3 */
		cartbank[slot][0]=cartbank[slot][4]=cartbank[slot][6]=2&cartmask[slot];
		cartbank[slot][1]=cartbank[slot][5]=cartbank[slot][7]=3&cartmask[slot];
	}
	
	refresh_cb(slot);
}

static void mapinit_crossblaim(int slot)
{
	int i;
	
	/* 2 pages of 16KB at $4000-$bfff */
	cartbank[slot][2]=0; cartbank[slot][3]=1; /* bank 0 fixed to page 1 */
	cartbank[slot][0]=cartbank[slot][4]=cartbank[slot][6]=2&cartmask[slot];
	cartbank[slot][1]=cartbank[slot][5]=cartbank[slot][7]=3&cartmask[slot];
	
	for (i=0;i<8;i++) mapwrite_std[slot+1][i]=mapwrite_crossblaim;
}



/* --- Matra INK (single game) */
static _am29f* ink_chip[2]={NULL,NULL};

u8 __fastcall mapread_matraink(u16 a)
{
	return am29f_read(ink_chip[~pslot>>(a>>13&6)&1],a);
}

void __fastcall mapwrite_matraink(u16 a,u8 v)
{
	am29f_write(ink_chip[~pslot>>(a>>13&6)&1],a,v);
}

static void mapclean_matraink(int slot)
{
	am29f_clean(ink_chip[slot]);
	ink_chip[slot]=NULL;
}

/* state			size
amd mode			2
amd inc				1

==				3
*/
#define STATE_VERSION_INK	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_INK		3

static int __fastcall mapgetstatesize_matraink(int slot)
{
	return STATE_SIZE_INK;
}

static void __fastcall mapsavestate_matraink(int slot,u8** s)
{
	STATE_SAVE_2(ink_chip[slot]->mode);
	STATE_SAVE_1(ink_chip[slot]->inc);
}

static void __fastcall maploadstatecur_matraink(int slot,u8** s)
{
	STATE_LOAD_2(ink_chip[slot]->mode);
	STATE_LOAD_1(ink_chip[slot]->inc);
}

static int __fastcall maploadstate_matraink(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_INK:
			maploadstatecur_matraink(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_matraink(int slot)
{
	int i;
	
	if (mapclean[slot]==mapclean_nothing) {
		ink_chip[slot]=am29f_init(TRUE);
		mapclean[slot]=mapclean_matraink;
		
		/* upload to flash */
		for (i=0;i<0x40;i++) memcpy(ink_chip[slot]->data+0x2000*i,cart[slot][i],0x2000);
	}
	
	for (i=0;i<8;i++) ink_chip[slot]->unprotect[i]=TRUE;
	ink_chip[slot]->readsector=0;
	ink_chip[slot]->writesector=0x40000;
	am29f_reset(ink_chip[slot]);
	
	mapgetstatesize[slot]=mapgetstatesize_matraink;
	mapsavestate[slot]=mapsavestate_matraink;
	maploadstatecur[slot]=maploadstatecur_matraink;
	maploadstate[slot]=maploadstate_matraink;
	
	for (i=0;i<8;i++) {
		mapwrite_std[slot+1][i]=mapwrite_matraink;
		mapread_std[slot+1][i]=mapread_matraink;
	}
}



/* --- MicroCabin Harry Fox (single game: Harry Fox - Yuki no Maou hen) */
void __fastcall mapwrite_harryfox(u16 a,u8 v)
{
	/* page 1 switchable at $6xxx, page 2 switchable at $7xxx, bit 0:select 32KB ROM */
	int slot=~pslot>>2&1;
	int p=a>>11&2;
	int b=((v<<2&4)|p)&cartmask[slot];
	p+=2;
	
	cartbank[slot][p++]=b++;
	cartbank[slot][p]=b;
	refresh_cb(slot);
}

static void mapinit_harryfox(int slot)
{
	int i;
	
	/* 2 pages of 16KB at $4000-$bfff, init to 0,1 */
	for (i=0;i<4;i++) cartbank[slot][i+2]=i&cartmask[slot];
	
	mapwrite_std[slot+1][3]=mapwrite_harryfox;
}



/* --- Irem R-Type (single game) */
void __fastcall mapwrite_rtype(u16 a,u8 v)
{
	if (a&0x1000) {
		/* 2nd page switchable at $7000-$7fff, bit 4 selects romchip */
		int slot=~pslot>>2&1;
		int mask=(v&0x10)+0x1e;
		
		cartbank[slot][2]=mask;		cartbank[slot][3]=mask|1;
		cartbank[slot][4]=v<<1&mask;	cartbank[slot][5]=(v<<1&mask)|1;
		refresh_cb(slot);
	}
}

static void mapinit_rtype(int slot)
{
	/* 2 romchips: one 256KB, one 128KB. 2 pages of 16KB at $4000-$bfff, 1st page fixed to last bank of current romchip */
	cartbank[slot][2]=0x1e;	cartbank[slot][3]=0x1f;
	cartbank[slot][4]=0;	cartbank[slot][5]=1;
	
	mapwrite_std[slot+1][3]=mapwrite_rtype;
}



/* --- Sony Playball (single game)

NOTE: If samples are used, it can cause reverse and speed up problems, unable to fix it for now.
 (eg. reverse -> instant replay, can give different playback due to sample/buffer offset)

*/
static _sample* playball_sample[2]={NULL,NULL};

u8 __fastcall mapread_playball(u16 a)
{
	int slot=~pslot>>4&1;
	
	/* register at $bfff, bit 0: 1=ready, 0=busy */
	if (a==0xbfff) return playball_sample[slot]->playing^0xff;
	
	return cartmap[slot][5][a&0x1fff];
}

void __fastcall mapwrite_playball(u16 a,u8 v)
{
	if (a==0xbfff&&v<15) {
		/* register at $bfff, play sample if ready */
		int slot=~pslot>>4&1;
		if (!playball_sample[slot]->playing) sample_play(playball_sample[slot],slot,v);
	}
}

static void mapstream_playball(int slot,signed short* stream,int len)
{
	sample_stream(playball_sample[slot],slot,stream,len);
}

static int __fastcall mapgetstatesize_playball(int slot) { return sample_state_get_size(); }
static void __fastcall mapsavestate_playball(int slot,u8** s) { sample_state_save(playball_sample[slot],s); }
static void __fastcall maploadstatecur_playball(int slot,u8** s) { sample_state_load_cur(playball_sample[slot],s); }
static int __fastcall maploadstate_playball(int slot,int v,u8** s) { return sample_state_load(playball_sample[slot],v,s); }

static void mapclean_playball(int slot)
{
	sample_stop(playball_sample[slot]);
	sample_clean(playball_sample[slot]);
	playball_sample[slot]=NULL;
}

static void mapinit_playball(int slot)
{
	if (mapclean[slot]==mapclean_nothing) {
		playball_sample[slot]=sample_init();
		MEM_CLEAN(file->name);
		if (mapper_get_file(slot)) file_setfile(NULL,mapper_get_file(slot),NULL,NULL);
		sample_load(playball_sample[slot],slot,file->name);
		
		mapclean[slot]=mapclean_playball;
		mapgetstatesize[slot]=mapgetstatesize_playball;
		mapsavestate[slot]=mapsavestate_playball;
		maploadstatecur[slot]=maploadstatecur_playball;
		maploadstate[slot]=maploadstate_playball;
		mapstream[slot]=mapstream_playball;
	}
	
	mapinit_4000(slot); /* mapper_simple.c */
	mapwrite_std[slot+1][5]=mapwrite_playball;
	mapread_std[slot+1][5]=mapread_playball;
}
